/* (The MIT License)
Copyright (c) 2006 Adam Bennett (cruxic@gmail.com)
 
Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:
 
The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
 
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
 */
package adamb.util;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;

public class FileInsert
{
	private byte[] chunk;
	
	public FileInsert(int readChunkSize)
	{
		assert readChunkSize > 0;
		chunk = new byte[readChunkSize];
	}
	
	/**
	 Insert, append, overwrite, or delete an interval of the given file.
	 
	 The simplest way to view this operation is:
	 1) the specified interval is deleted from the file (leaving no holes)
	 2) the new data is inserted at the start of the interval (extending the file length as needed)
	 The algorithm used is more efficient, however.
	 
	 All bytes on the interval [from, to) will be lost (unless from == to in which case no data will be lost).
	 
	 @param file an open file.  The file will <b>not</b> be closed.
	 @param from the file byte position to start insertion at.  Must be <= to the file length. (Equal implies an append)
	 @param to file position of the last byte to overwrite (exclusive).  If from == to then the operation will be purely an insert (no deleted data).  Must not be greater than the file size.
	 @param newData byte array containing the new data
	 @param offset the offset within newData to begin reading from
	 @param len the number of bytes to read from newData.  Pass 0 to delete the file interval
	 */
	public void insert(RandomAccessFile file, long from, long to, byte[] newData, int offset, int len)
	throws IOException
	{
		/*make sure the given interval is acceptable.
		 
			It is important to make sure that everything is in order before we
			start modifying the file because the changes being made are not reversable!
		 */
		long fileLen = file.length();
		if
			(
			//unacceptable interval?
			from < 0 || from > to
			//from and to cant be greater than the file length!
			|| from > fileLen || to > fileLen
			)
			throw new IllegalArgumentException("Invalid insertion interval!: from=" + from + " to=" + to + " (file size=" + fileLen + ")");
		
		//make sure the given data is valid ahead of time
		if (len > 0)
		{
			//the following will throw ArrayIndexOutOfBoundsException if invalid
			byte b = newData[offset];
			b = newData[offset + len - 1];
		}
		
		//the number of bytes to delete
		long intervalSize = to - from;
		
		//the amount the file must be extended (positive) or shrunken (negative)
		long excess = len - intervalSize;
		
		//case 1: file must grow
		if (excess > 0)
		{
			int nRead;
			boolean stop = false;
			int chunkSize = chunk.length;
			long lastPos = fileLen;
			
			while (!stop)
			{
				lastPos -= chunkSize;
				
				//will this be the last chunk
				if (lastPos <= to)
				{
					chunkSize = (int)((lastPos + chunkSize) - to);
					lastPos = to;
					stop = true;
				}
				
				file.seek(lastPos);
				
				//read the entire chunk
				nRead = readCompletely(file, chunk, chunkSize);
				if (nRead != chunkSize)
					throw new IOException("Unexpected read shortage: " + nRead + " bytes instead of " + chunkSize + '!');
				
				file.seek(lastPos + excess);
				file.write(chunk, 0, chunkSize);
			}
		}
		//case 2: file must shrink
		else if (excess < 0)
		{
			long lastPos = to;
			int nRead;
			while (true)
			{
				file.seek(lastPos);
				nRead = file.read(chunk);
				//are we done?
				if (nRead == -1)
					break;
				file.seek(lastPos + excess);  //since excess is negative this will seek backwards from lastPos
				file.write(chunk, 0, nRead);
				lastPos += nRead;
			}
			
			//truncate the file
			file.setLength(fileLen + excess);  //since excess is negative the file will be truncated
		}
		//case 3: file size does not change (no special action necessary)
		
		//finally write in the new data
		file.seek(from);
		file.write(newData, offset, len);
	}
	
	
	/**block until the given byte array can be filled from the stream
	 @return the number bytes actually read.  This will only be less than the length
	 of the byte array if the end of the stream is reached.
	 */
	private int readCompletely(RandomAccessFile is, byte[] bytes, int numToRead)
	throws IOException
	{
		int nRead = is.read(bytes, 0, numToRead);
		while (nRead < numToRead)
		{
			if (nRead == -1)
				return 0;
			else
				nRead += is.read(bytes, nRead, numToRead - nRead);
		}
		
		return nRead;
	}
	
}