package us.lakora.brawl.rateconv;

/*****************************************************************************
 *   Copyright 2012 Isaac Schemm
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   
 *   You may obtain a copy of the Apache License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 ****************************************************************************/

import java.awt.Color;
import java.awt.Dimension;
import java.io.File;
import java.io.IOException;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.UIManager;

import us.lakora.brawl.rateconv.music.*;

public class Main {

	public static void main(String[] args) {

		try {
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		} catch (Exception e) {
			
		}
		
//		try {
//			Class.forName("java.nio");
//		} catch (ClassNotFoundException e3) {
//			JOptionPane.showMessageDialog(null, "This program requires Java 7 or newer.", Shared.PROGRAM_TITLE, JOptionPane.ERROR_MESSAGE);
//			System.exit(1);
//		}
		
		File tmp_dir = new File(Shared.TMP_DIR);
		if (!tmp_dir.isDirectory()) {
			if (tmp_dir.exists()) {
				Shared.errorDialog("There is a file named \"tmp\". Please move this file so the program can create a \"tmp\" directory.");
				System.exit(1);
			} else {
				tmp_dir.mkdir();
			}
		}

		errorCheck();

		File[] inputFiles = null;
		boolean cmdLineOptions = false;

		OptionSet opts = new OptionSet();
		ArrayList<File> foundFiles = new ArrayList<File>(args.length);
		ArrayList<String> notFoundFiles = new ArrayList<String>();

		// This code processes the arguments.
		for (int i=0; i<args.length; i++) {
			if (args[i].charAt(0) == '-') {
				cmdLineOptions = true; // Set to true so the options dialog doesn't appear
				char first = args[i].charAt(1);
				// Arguments defined here
				if ((first == 'm') && (args[i].length() == 2)) {
					opts.setConvertToMono(true);
				} else if ((first == 'd') && (args[i].length() == 2)) {
					i++; // Skip to next argument - it should be a number
					try {
						int num = Integer.parseInt(args[i]);
						opts.setDefaultRate((int)num);
					} catch (NumberFormatException e) {
						System.err.println(args[i] + " is not a number.");
					}
				} else if ((first == 'r') && (args[i].length() == 2)) {
					i++; // Skip to next argument - it should be a number
					try {
						double num = Double.parseDouble(args[i]);
						if (num % 1 != 0.0) { // If the number is not an integer
							opts.setConvertRate(OptionSet.RATE_RELATIVE);
							opts.setRateFactor(num);
						} else { // If it is an integer
							opts.setConvertRate(OptionSet.RATE_ABSOLUTE);
							opts.setDefaultRate((int)num);
						}
					} catch (NumberFormatException e) {
						System.out.println(args[i] + " is not a number.");
					}
				} else if ((first == 'a') && (args[i].length() == 2)) {
					i++; // Skip to next argument - it should be a number
					try {
						double num = Double.parseDouble(args[i]);
						opts.setAmpAmount(num); // Will not be set if Double.parseDouble throws an exception
					} catch (NumberFormatException e) {
						System.err.println(args[i] + " is not a number.");
					}
				} else if (args[i].equals("min")) {
					i++; // Skip to next argument - it should be a number
					try {
						int num = Integer.parseInt(args[i]);
						opts.setMinimumRate(num); // Will not be set if Integer.parseInt throws an exception
					} catch (NumberFormatException e) {
						System.err.println(args[i] + " is not a number.");
					}
				} else if (args[i].equals("-loop")) {
					opts.setLoopWav(true);
				} else if (args[i].equals("-noloop")) {
					opts.setLoopWav(false);
				} else if ((first == 'l') && (args[i].length() == 2)) {
					opts.setLoopWav(true); // If you define a loop file, it is implied that it should be used
					i++; // Next argument - it should be a filename
					File loopFile = new File(args[i]);
					if (loopFile.isFile()) {
						opts.setLoopTxt(args[i]);
					} else {
						System.err.println(args[i] + " does not exist.");
					}
				} else if (args[i].equals("-noopts")) {
					// Do nothing except set cmdLineOptions to true (which is done above) so the options dialog doesn't show up
				} else if (args[i].equals("--help")) {
					System.out.println(Shared.HELPMESSAGE);
					System.exit(0);
				} else {
					System.err.println("Unrecognized argument: " + args[i] + "\n\n" + Shared.HELPMESSAGE);
					System.exit(1);
				}
			} else {
				// For arguments that are input filenames
				File f = new File(args[i]);
				if (f.exists()) {
					foundFiles.add(f);
				} else {
					notFoundFiles.add(args[i]);
				}
			}
		}
		if (notFoundFiles.size() > 0) {
			System.out.println("Not found: " + notFoundFiles);
		}
		if (foundFiles.size() > 0) {
			File[] tmp = new File[foundFiles.size()];
			foundFiles.toArray(tmp);
			inputFiles = tmp; // Put the found files into the array
		}

		// Run this if there are no options on the command line (empty or only filenames)
		if (!cmdLineOptions) {
			OptionDialog od = new OptionDialog(opts);
			if (!od.ask()) {
				System.exit(0);
			}
		}

		int vgmRate; // the rate to convert VGM files to
		if (opts.getConvertRate() == OptionSet.RATE_ABSOLUTE) {
			vgmRate = (int)opts.getDefaultRate();
		} else {
			vgmRate = 44100;
		}

		// Run this if there are no filenames on the command line (only args or empty)
		if ((inputFiles == null) || (inputFiles.length == 0)) {
			JFileChooser chooser = new JFileChooser();
			try {
				// Create a File object containing the canonical path of the
				// desired directory
				File f = new File(new File(".").getCanonicalPath());

				// Set the current directory
				chooser.setCurrentDirectory(f);
			} catch (IOException e1) {
				e1.printStackTrace();
				// We can still continue the program. Anyway, this error should never happen.
			}

			ExtensionFileFilter f1 = new ExtensionFileFilter("Wii audio (.brstm)", new String[] { "BRSTM" });
			ExtensionFileFilter f2 = new ExtensionFileFilter("vgmstream-supported formats", FileExtensions.VGMSTREAM);
			ExtensionFileFilter f3 = new ExtensionFileFilter("SoX-supported formats (.mp3, .wav, .flac...)", FileExtensions.SOX);
			ExtensionFileFilter f5 = new ExtensionFileFilter("VGM music log", new String[] { "VGM", "VGZ" });
			ExtensionFileFilter all = new ExtensionFileFilter("All of the above", Shared.concat(
					FileExtensions.SOX,
					FileExtensions.VGMSTREAM,
					new String[] { "VGM", "VGZ" }));
			chooser.addChoosableFileFilter(f1);
			chooser.addChoosableFileFilter(f2);
			chooser.addChoosableFileFilter(f3);
			chooser.addChoosableFileFilter(f5);
			chooser.addChoosableFileFilter(all);
			chooser.setAcceptAllFileFilterUsed(true);

			chooser.setFileFilter(all);
			chooser.setMultiSelectionEnabled(true);

			if ((!Shared.OUTPUT_DIR.exists()) && (!Shared.OUTPUT_DIR.mkdir())) {
				Shared.errorDialog("Could not create directory \"" + Shared.OUTPUT_DIR + "\".");
				System.exit(1);
			}

			int returnVal = chooser.showOpenDialog(null);
			if (returnVal != JFileChooser.APPROVE_OPTION) {
				System.exit(1);
			}
			inputFiles = chooser.getSelectedFiles();
		}

		LoopTxtReader loopReader;
		try {
			loopReader = new LoopTxtReader(opts.getLoopTxt());
		} catch (IOException e2) {
			if (opts.isLoopWav()) {
				Shared.errorDialog("Could not read " + opts.getLoopTxt() + ".\nNon-looping formats (such as WAV or OGG) will be looped start to end.");
			} else {
				Shared.errorDialog("Could not read " + opts.getLoopTxt() + ".\nNon-looping formats (such as WAV or OGG) will not be looped.");
			}
			loopReader = new LoopTxtReader(); // Make an empty LoopTxtReader that returns the defaults for all songs
		}

		// The setLoops function takes a long time on Wine (multiple .exe calls) so let's display a box.
		// This is also good because we might need to convert MP3s.
		JFrame noteFrame = new JFrame();
		ProgressBox note = new ProgressBox();
		note.changeMessage("Please wait while the file information is read.");
		noteFrame.setTitle(Shared.PROGRAM_TITLE);
		noteFrame.add(note);
		noteFrame.pack();
		noteFrame.setVisible(true);

		noteFrame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE); // This is OK because we don't have anything to clean up yet

		// This array will be used for input files
		Music[] inputMusics = new Music[inputFiles.length];
		// A list of files that could not be converted
		List<File> couldNotConvert = new ArrayList<File>();
		// The current file
		File file = null;
		try {
			for (int i=0; i<inputMusics.length; i++) {
				file = inputFiles[i];
				note.changeStatus(file.getName(), Color.blue);
				String name = file.getCanonicalPath();
				String ext = name.substring(name.lastIndexOf('.')+1).toLowerCase();
				if (Arrays.asList(FileExtensions.SOX).contains(ext)) {
					Wav w = new Wav(inputFiles[i]);
					loopReader.setLoops(w, opts.isLoopWav()); // Will run setLoop as appropriate
					inputMusics[i] = w;
				} else if ("vgm".equals(ext) || "vgz".equals(ext)) {
					inputMusics[i] = new VGM(inputFiles[i], vgmRate);
				} else {
					// handled by vgmstream
					inputMusics[i] = new Vgmstream(inputFiles[i]);
				}
			}
		} catch (IOException e) {
			e.printStackTrace();
			Shared.errorDialog("I/O error: " + file);
			couldNotConvert.add(file);
		} catch (MusicFileException e) {
			e.printStackTrace();
			Shared.errorDialog(e.getMessage());
			couldNotConvert.add(file);
		} catch (InterruptedException e) {
			// stay quiet on this one
			couldNotConvert.add(file);
		}

		noteFrame.setVisible(false);
		noteFrame.dispose();

		// Start the conversion!
		
		ProgressBox jld = new ProgressBox();
		JFrame frame = new JFrame(Shared.PROGRAM_TITLE);
		frame.setContentPane(jld);
		frame.setSize(300, 75);
		frame.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
		AL frameListener = new AL();
		frame.addWindowListener(frameListener);
		frame.setVisible(true);

		List<File> filesMade = new ArrayList<File>(); // A list of files converted
		int readableBrstms = inputMusics.length;
		Music[] newInputArray = new Music[readableBrstms];
		int i = 0; // newInputArray counter
		for (int j=0; j<inputMusics.length; j++) {
			if (inputMusics[j] != null) { // Don't copy null elements
				newInputArray[i] = inputMusics[j];
				i++;
			}
		}
		inputMusics = newInputArray;

		File[] outputFiles = new File[readableBrstms];

		boolean cont = true;
		for (i=0; (i<inputMusics.length) && (cont); i++) if (inputMusics[i] != null) {
			Music input = inputMusics[i];

			outputFiles[i] = new File(Shared.OUTPUT_DIR + File.separator + input.getOutputName(opts.getOutputFormat().getExt()));
			
			if (input.create(jld, opts, outputFiles[i])) {
				filesMade.add(outputFiles[i]);
			} else {
				couldNotConvert.add(input.getInput());
			}

			if (frameListener.wasPressed()) {
				cont = false;
			}
		}
		frame.dispose();

		StringBuilder filesMadeString = new StringBuilder("Files made:\n");
		int filesMadeSize = filesMade.size();
		for (int j=0; j<filesMadeSize; j++) {
			filesMadeString.append(filesMade.get(j));
			if (j < filesMadeSize-1) {
				filesMadeString.append(", ");
			} else {
				filesMadeString.append("\n");
			}
		}
		String cannotConvert;
		if (couldNotConvert.size() > 0) {
			StringBuilder sb = new StringBuilder();
			sb.append("Files that could not be read:\n");
			sb.append(couldNotConvert.get(0));
			for (int j=1; j<couldNotConvert.size(); j++) {
				sb.append(", ");
				sb.append(couldNotConvert.get(i));
			}
			sb.append("\n");
			cannotConvert = sb.toString();
		} else {
			cannotConvert = "";
		}

		// Show "finished" dialog
		JTextArea area = new JTextArea(filesMadeString + cannotConvert);
		area.setLineWrap(true);
		area.setWrapStyleWord(true);
		JScrollPane scrollPane = new JScrollPane();
		scrollPane.getViewport().setView(area);
		scrollPane.setPreferredSize(new Dimension(300, 200));

		JFrame finished_frame = new JFrame("Finished - " + Shared.PROGRAM_TITLE);
		finished_frame.setVisible(true);
		JOptionPane.showMessageDialog(finished_frame, scrollPane);
		System.exit(0);
	}

	private static void errorCheck() {
		// If there are certain files that don't exist, use DialogList to tell the user about them.
		List<String> errors = new ArrayList<String>();
		if (
				(!new File(Shared.DSPADPCM_DIR + File.separator + "dsptool.dll").exists()) ||
				(!new File(Shared.DSPADPCM_DIR + File.separator + "hio2.dll").exists()) ||
				(!new File(Shared.DSPADPCM_DIR + File.separator + "soundfile.dll").exists())) {
			errors.add("Some of the DLL files required for dspadpcm.exe are not present in the \"" + Shared.DSPADPCM_DIR + "\" folder.");
		}
		if (!new File(Shared.VGMSTREAM_DIR + File.separator + "test.exe").exists()) {
			errors.add("test.exe is not present in the \"" + Shared.VGMSTREAM_DIR + "\" folder.\nYou can get test.exe from:\nhttp://hcs64.com/vgmstream.html");
		}
		if (!new File(Shared.SOX_DIR + File.separator + "sox.exe").exists()) {
			errors.add("sox.exe is not present in the \"" + Shared.SOX_DIR + "\" folder.\nYou can download SoX from: http://sourceforge.net/projects/sox/files/sox-win/\n(Get the .zip file, not the installer.)");
		}
		// Continue only if there are no errors.
		if (errors.size() > 0) {
			String errorString;
			if (errors.size() < 3) {
				errorString = errors.get(0);
				for (int i=1; i<errors.size(); i++) errorString += "\n" + errors.get(i);
			} else {
				errorString = "There are more than 3 file-not-found errors.\nMake sure that the Rate Converter JAR file is in the right directory (with the \"tools\" and \"sox\" folders.)";
			}
			JOptionPane.showMessageDialog(null, errorString, Shared.PROGRAM_TITLE, JOptionPane.ERROR_MESSAGE);
			System.exit(1);
		}
		
		if (!new File(Shared.DSPADPCM_DIR + File.separator + "dspadpcm.exe").exists()) {
			errors.add("dspadpcm.exe is not present in the \"" + Shared.DSPADPCM_DIR + "\" folder.\n" +
					"You will not be able to create .brstm files without it.\n" +
					"This file is part of Nintendo's SDK, but you might be able to find a copy packaged with another BRSTM-related program.");
		}
		if (!new File(Shared.SOX_DIR + File.separator + "libmad.dll").exists()) {
			JOptionPane.showMessageDialog(null, "libmad.dll is not present in the \"" + Shared.SOX_DIR + "\" folder.\nYou will not be able to decode MP3 files without it.");
		}
	}

}
