package us.lakora.brawl.rateconv.music;

import java.awt.Color;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;

import adamb.vorbis.CommentField;
import adamb.vorbis.CommentUpdater;
import adamb.vorbis.VorbisCommentHeader;
import adamb.vorbis.VorbisIO;

import us.lakora.brawl.rateconv.MiniCopyFile;
import us.lakora.brawl.rateconv.OptionSet;
import us.lakora.brawl.rateconv.ProgressBox;
import us.lakora.brawl.rateconv.Shared;
import us.lakora.brawl.revb.RevB;
import us.lakora.brawl.revb.RevB.RevBException;


/*****************************************************************************
 *   Copyright 2012 Isaac Schemm
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   
 *   You may obtain a copy of the Apache License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 ****************************************************************************/

public abstract class Music {
	
	protected File input;
	protected File temporaryWavFile;

	protected int sampleRate;
	protected int loopStart;
	protected int length;
	protected int channels;

	/**
	 * Samples per block in a BRSTM file for Brawl = 14336.
	 */
	public static final int SAMPLES_PER_BLOCK = 14336;
	
	protected Music(File input) {
		this.input = input;
		this.loopStart = -1; // not looping by default
		
		// set to invalid values; these should be changed by the subclass
		this.length = -1;
		this.sampleRate = -1;
		this.channels = -1;
	}
	
	/**
	 * Either creates the temporary WAV file, or sets it to be the same as the
	 * input file. The implementing class is responsible for deleting this
	 * temporary file when the program quits (using deleteOnExit() is recommended.
	 * @param progressBox Can be null, or a ProgressBox that can be updated by this method and other methods that call it.
	 * @throws IOException
	 * @throws InterruptedException
	 * @throws MusicFileException
	 */
	
	public abstract void createTemporaryWavFile(ProgressBox progressBox) throws IOException, InterruptedException, MusicFileException;

	/**
	 * Returns the original input file.
	 */
	public File getInput() {
		return input;
	}
	/**
	 * Returns the original sample rate (immutable).
	 */
	public int getSampleRate() {
		return sampleRate;
	}
	/**
	 * Returns the loop start in samples, or -1 if the song does not loop. Can be changed by a subclass; e.g. NonLoopingMusic, which has a setLoop method.
	 */
	public int getLoopStart() {
		return loopStart;
	}
	/**
	 * Returns the length, or loop end, in samples. Can be changed by a subclass; e.g. NonLoopingMusic, which has a setLoop method.
	 */
	public int getLength() {
		return length;
	}
	/**
	 * Returns the number of channels in the input file.
	 */
	public int getChannels() {
		return channels;
	}

//	/**
//	 * Sets the extension of the output file. Must be either BRSTM or a SoX-supported format.
//	 * @throws MusicFileException 
//	 */
//	protected void setOutputExt(String outputExt) throws MusicFileException {
//		outputExt = outputExt.toLowerCase();
//		if (outputExt.equals("brstm")
//				|| outputExt.equals("mp3")
//				|| Arrays.asList(FileExtensions.SOX).contains(outputExt)) {
//			this.outputExt = outputExt;
//		} else {
//			throw new MusicFileException("Output extension must be either BRSTM or a SoX-supported format.");
//		}
//	}

	/**
	 * A convenience method. Equivalent to (getLoopStart >= 0).
	 */
	public boolean isLooping() {
		return (getLoopStart() >= 0);
	}

	/**
	 * Returns the same value as getName().
	 */
	public String toString() {
		return getName();
	}
	
	/**
	 * Returns the canonical path of the input file, surrounded by quotes.
	 */
	public String quotedName() {
		try {
			return "\"" + input.getCanonicalPath() + "\"";
		} catch (IOException e) {
			// I've never had this happen
			e.printStackTrace();
			return "\"" + input.getAbsolutePath() + "\"";
		}
	}
	
	/**
	 * Returns the result of {@link File#getCanonicalPath()} on the input file.
	 * @return the canonical path of the file.
	 * @throws IOException
	 */
	public String getCanonicalPath() throws IOException {
		return input.getCanonicalPath();
	}
	
	/**
	 * Returns the result of {@link File#getName()} on the input file.
	 * @return the name path of the file.
	 * @throws IOException
	 */
	public String getName() {
		return input.getName();
	}
	
	/**
	 * Returns the result of getName() with the extension replaced with brstm.
	 */
	public String getOutputName() {
		return getOutputName("brstm");
	}
	
	/**
	 * Returns the result of getName() with the extension replaced with the given one.
	 */
	public String getOutputName(String extension) {
		return input.getName().substring(0, input.getName().lastIndexOf(".")+1) + extension;
	}

	public boolean create(ProgressBox progressBox, OptionSet opts, File outputFile) {
		return create(progressBox, opts, outputFile, "brstm");
	}
	
	/**
	 * Creates a BRSTM from the input file.
	 * @param progressBox Can be null, or a ProgressBox that can be updated by this method and other methods that call it.
	 * @param opts An OptionSet to read parameters from.
	 * @param outputFile The file to output to.
	 * @return true if the conversion was successful, false otherwise.
	 */
	public boolean create(ProgressBox progressBox, OptionSet opts, File outputFile, String extension) {
		int newRate;
		if (opts.getConvertRate() == OptionSet.RATE_RELATIVE) {
			newRate = (int)Math.round(opts.getRateFactor() * this.getSampleRate());
			if (newRate < opts.getMinimumRate()) {
				newRate = opts.getMinimumRate(); // Change to the minimum rate - this only needs to be checked if the rate is defined as a fraction
			}
		} else if (opts.getConvertRate() == OptionSet.RATE_ABSOLUTE) {
			newRate = opts.getDefaultRate();
		} else {
			newRate = this.getSampleRate();
		}
		
		if (progressBox != null) {
			progressBox.changeStatus("Converting " + input.getName() + "...");
			
			if (newRate >= this.getSampleRate()) {
				newRate = this.getSampleRate(); // This will keep the below code from changing the sample rate for this Music.
			}
	
			int channels = this.getChannels();
	
			String message = "";
	
			if (this.isLooping()) {
				message = message + "loop " + this.getLoopStart() + "-" + this.getLength() + ", ";
			}
			if (opts.isConvertToMono() && (channels > 1)) {
				message = message + "to mono, ";
			}
			if (opts.getConvertRate() != OptionSet.RATE_NO_CONVERSION && this.getSampleRate() != newRate) {
				message = message + "new rate: " + newRate + ", ";
			}
			if (opts.getAmpType() == OptionSet.AMPLITUDE_TYPE) {
				message = message + "amp: " + opts.getAmpAmount() + "x, ";
			} else if (opts.getAmpType() == OptionSet.DB_TYPE) {
				message = message + "amp: " + opts.getAmpAmount() + " dB, ";
			} else if (opts.getAmpType() == OptionSet.MAX_AMP_WITHOUT_CLIPPING) {
				message = message + "max amp w/o clipping, ";
			}
			if (message.length() > 2) { // do this if the message is not empty
				message = message.substring(0, message.length()-2); // remove last comma and space
				progressBox.changeMessage(message, Color.blue);
			}
		}

		try { // this is where the magic happens!
			File[] wavFiles = this.extractWAV(progressBox);

			/**
			 * WAV files after stereo/mono conversion.
			 */
			File[] step1wav;
			if (channels == 1) { // already mono
				step1wav = new File[1];
				step1wav[0] = wavFiles[0];
			} else if (opts.isConvertToMono()) { // not mono; converting to mono
				step1wav = new File[1];
				step1wav[0] = combineToMono(wavFiles); // this function uses only the first two channels
			} else { // not mono; converting to stereo
				step1wav = new File[2]; // Drop any other audio channels (for NSMBWii, Klonoa, Wii Sports Resort, etc.)
				step1wav[0] = wavFiles[0];
				step1wav[1] = wavFiles[1];
				for (int j=2; j<wavFiles.length; j++) {
					wavFiles[j].delete();
				}
			}

			/**
			 * WAV files after sample rate conversion.
			 */
			File[] step2wav;
			if (opts.getConvertRate() != OptionSet.RATE_NO_CONVERSION && newRate != this.getSampleRate()) {
				step2wav = new File[step1wav.length];
				for (int j=0; j<step2wav.length; j++) {
					step2wav[j] = new File(Shared.TMP_DIR + File.separator + "step2_" + 0 + "_" + j + ".wav");
					String[] cmdLine = {Shared.SOX_DIR + File.separator + "sox.exe",
							step1wav[j].getCanonicalPath(),
							"-r", newRate+"",
							step2wav[j].getCanonicalPath()};
					Process rate = Runtime.getRuntime().exec(cmdLine);
					rate.waitFor();
					step1wav[j].delete();
				}
			} else {
				step2wav = step1wav;
			}

			/**
			 * WAV files after amplification.
			 */
			File[] step3wav;
			boolean amplify = true;
			String[] soxOpts = new String[2];
			if (opts.getAmpType() == OptionSet.NO_AMP) {
				amplify = false;
			} else if (opts.getAmpType() == OptionSet.AMPLITUDE_TYPE) {
				soxOpts[0] = "vol";
				soxOpts[1] = opts.getAmpAmount() + "amplitude";
			} else if (opts.getAmpType() == OptionSet.DB_TYPE) {
				soxOpts[0] = "vol";
				soxOpts[1] = opts.getAmpAmount() + "dB";
			} else if (opts.getAmpType() == OptionSet.MAX_AMP_WITHOUT_CLIPPING) {
				soxOpts[0] = "gain";
				soxOpts[1] = "-n";
			}
			if (amplify) {
				step3wav = new File[step2wav.length];
				for (int j=0; j<step3wav.length; j++) {
					step3wav[j] = new File(Shared.TMP_DIR + File.separator + "step3_" + 0 + "_" + j + ".wav");
					String[] cmdLine = {Shared.SOX_DIR + File.separator + "sox.exe",
							step2wav[j].getCanonicalPath(), step3wav[j].getCanonicalPath(),
							soxOpts[0], soxOpts[1]};
					Process amp = Runtime.getRuntime().exec(cmdLine);
					amp.waitFor();
					step2wav[j].delete();
				}
			} else {
				step3wav = step2wav;
			}
			// make the new BRSTMs
			boolean res;
			if (opts.getOutputFormat().getExt().equals("brstm")) {
				res = this.buildBrstm(step3wav, outputFile, newRate);
			} else {
				res = this.buildWav(step3wav, outputFile, opts);
			}

			for (int j=0; j<step3wav.length; j++) {
				step3wav[j].delete();
			}
			return res;
			
		} catch (Exception e) {
			e.printStackTrace();
			Shared.errorDialog(e.getClass().getName() + ": " + e.getMessage());
			return false;
		}
	}
	
	private static File combineToMono(File[] wavFiles) throws IOException, InterruptedException {
		File tmpwav = new File(Shared.TMP_DIR + File.separator + "tmp.wav");
		File outwav = new File(Shared.TMP_DIR + File.separator + "out.wav");
		if (wavFiles.length != 1) {
			if (wavFiles.length > 2) System.out.println("This stream has more than two channels. Only using first two.");
			final int CHANNELS = 2;
			File[] amplifiedWavFiles = new File[CHANNELS];
			for (int i=0; i<CHANNELS; i++) {
				amplifiedWavFiles[i] = new File(Shared.TMP_DIR + File.separator + "amp" + i + ".wav");
				String[] cmdLine = {Shared.SOX_DIR + File.separator + "sox.exe", wavFiles[i].getCanonicalPath(), amplifiedWavFiles[i].getCanonicalPath(), "vol", (1.0 / wavFiles.length)+""};
				Process amplify = Runtime.getRuntime().exec(cmdLine);
				amplify.waitFor();
				wavFiles[i].delete();
			}
			for (int i=CHANNELS; i<wavFiles.length; i++) {
				wavFiles[i].delete();
			}
			List<String> arguments = new ArrayList<String>(3 + 3*CHANNELS);
			arguments.add(Shared.SOX_DIR + File.separator + "sox.exe");
			arguments.add("-m");
			for (int i=0; i<CHANNELS; i++) {
				arguments.add("-v");
				arguments.add("1.0");
				arguments.add(amplifiedWavFiles[i].getCanonicalPath());
			}
			arguments.add(tmpwav.getCanonicalPath());
			Process mix = new ProcessBuilder(arguments).start();
			mix.waitFor();
			for (int i=0; i<CHANNELS; i++) {
				amplifiedWavFiles[i].delete();
			}
			String[] args2 = {Shared.SOX_DIR + File.separator + "sox.exe", tmpwav.getCanonicalPath(), "-c", "1", outwav.getCanonicalPath()};
			Process toMono = Runtime.getRuntime().exec(args2);
			toMono.waitFor();
			tmpwav.delete();
		} else {
			System.out.println("Why are you running combineToMono() on a mono stream?");
			outwav = wavFiles[0];
		}
		return outwav;
	}

	private boolean buildWav(File[] outwav, File output, OptionSet opts) throws MusicFileException, IOException, InterruptedException {
		// making a looping file, such as .logg, will force loop trim to avoid putting a large amount of unused information in the file
		// this isn't done for .brstm files because dspadpcm should handle it.
		boolean loopTrim = opts.isLoopTrim() || opts.getOutputFormat().canLoop();
		String ext = opts.getOutputFormat().getExt();
		if (outwav.length == 2) {
			File[] temporaryFiles = new File[2];
			temporaryFiles[0] = new File(Shared.TMP_DIR + File.separator + "l.wav");
			temporaryFiles[1] = new File(Shared.TMP_DIR + File.separator + "r.wav");
			
			System.out.println("Converting mono files to stereo...");
			Process p1 = new ProcessBuilder(Shared.SOX_DIR + File.separator + "sox.exe",
					outwav[0].getPath(), "-c", "2",
					temporaryFiles[0].getPath(), "pan", "-1").start();
			Process p2 = new ProcessBuilder(Shared.SOX_DIR + File.separator + "sox.exe",
					outwav[1].getPath(), "-c", "2",
					temporaryFiles[1].getPath(), "pan", "1").start();
			p1.waitFor();
			p2.waitFor();

			if (p1.exitValue() != 0 || p2.exitValue() != 0) {
				throw new MusicFileException("Blending two WAV files to one was unsuccessful");
			}

			String outputPath = output.getPath();
			boolean logg = ext.equals("logg");
			if (logg) {
				outputPath = outputPath.replace(".logg", ".ogg");
			}
			Process p3;
			System.out.println("Combining stereo files...");
			if (isLooping() && loopTrim) {
				p3 = new ProcessBuilder(Shared.SOX_DIR + File.separator + "sox.exe",
						"-m",
						"-v", "1", temporaryFiles[0].getPath(),
						"-v", "1", temporaryFiles[1].getPath(),
						outputPath,
						"trim", "0", this.getLength()+"s").start();
			} else {
				p3 = new ProcessBuilder(Shared.SOX_DIR + File.separator + "sox.exe",
						"-m",
						"-v", "1", temporaryFiles[0].getPath(),
						"-v", "1", temporaryFiles[1].getPath(),
						outputPath).start();
			}
			p3.waitFor();
			
			temporaryFiles[0].delete();
			temporaryFiles[1].delete();

			if (p3.exitValue() != 0) {
				InputStream is = p3.getErrorStream();
				int read = is.read();
				while (read > 0) {
					System.out.print((char)read);
					read = is.read();
				}
				throw new MusicFileException("SoX error - check the console for details.");
			}

			if (logg) {
//				Files.move(new File(outputPath).toPath(), // .ogg file
//						output.toPath()); // .logg file - this is for Java 7
				new File(outputPath).renameTo(output);
				if (isLooping()) {
					VorbisIO.writeComments(output, new CommentUpdater() {
						public boolean updateComments(VorbisCommentHeader comments) {
							comments.fields.add(new CommentField("LOOPSTART", loopStart+""));
							comments.fields.add(new CommentField("LOOPLENGTH", (length-loopStart)+""));
							return true;
						}
					});
				}
			}

			return true;
		} else if (outwav.length == 1) {
			MiniCopyFile.copyFile(outwav[0], output);
			return true;
		} else {
			throw new MusicFileException("buildWav(...) can only take mono or stereo files");
		}
	}

	private static int dspadpcm(List<String> launch) throws IOException, InterruptedException {
		int exitValue = 1;
		Process p = new ProcessBuilder(launch).start();
		System.out.print("Launching: ");
		for (int i=0; i<launch.size(); i++) {
			System.out.print(launch.get(i) + " ");
		}
		System.out.println();
		Shared.progOut(p, "Output of dspadpcm:");
		p.waitFor();
		exitValue = p.exitValue();
		return exitValue;
	}
	
	private boolean buildBrstm(File[] outwav, File output, int newSampleRate) throws IOException, InterruptedException, MusicFileException {
		boolean keepGoing = true; // This is also the variable that is returned by the function - it tells the main function whether to include it in the list of ones that worked of the list of ones that didn't.
		double ratio = (double)newSampleRate / sampleRate;
		int newLoopStart = (int)(loopStart * ratio);
		int newLength = (int)(length * ratio);
		newLength--; // a fix for an odd "bug" in dspadpcm that seems to add an extra sample
		if (newLoopStart % SAMPLES_PER_BLOCK != 0) { // do this if the sample rate change has necessitated a change in the loop point.
			int addedSamples = SAMPLES_PER_BLOCK - (newLoopStart % SAMPLES_PER_BLOCK);
			newLoopStart += addedSamples;
			newLength += addedSamples;
		}
		for (int i=0; i<outwav.length; i++) {
			List<String> launch = new ArrayList<String>(5);
			launch.add(Shared.DSPADPCM_DIR + File.separator + "dspadpcm.exe");
			launch.add("-e");
			launch.add(outwav[i].getPath());
			launch.add(outwav[i].getPath().replaceAll("wav", "dsp"));

			if (isLooping()) {
				launch.add("-l" + newLoopStart + "-" + newLength);
			}

			int exitValue = dspadpcm(launch);
			outwav[i].delete();
			if (exitValue != 0) {
				throw new MusicFileException("dspadpcm.exe failed to convert " + this.getName());
			}
		}
		File[] outdsp = new File[outwav.length];
		for (int i=0; i<outwav.length; i++) {
			outdsp[i] = new File(outwav[i].getCanonicalPath().replaceAll("wav", "dsp"));
		}
		if (keepGoing) {
			System.out.println("Calling RevB.build(output, outdsp, outdsp.length, false, false)");
			try {
				RevB.build(output, outdsp, outdsp.length, false, false);
			} catch (RevBException e) {
				throw new MusicFileException(e.getMessage());
			}
		}
		for (int i=0; i<outwav.length; i++) {
			outdsp[i].delete();
			new File(outwav[i].getName().replaceAll("wav", "txt")).delete(); //dspadpcm.exe puts its txt files in the current directory
		}
		return keepGoing;
	}
	
	private File[] extractWAV(ProgressBox progressBox) throws IOException, InterruptedException, MusicFileException {
		if (temporaryWavFile == null) createTemporaryWavFile(progressBox);
		int channels = this.getChannels();
		File[] wavFiles = new File[channels];
		for (int i=0; i<channels; i++) {
			wavFiles[i] = new File(Shared.TMP_DIR + File.separator + "in" + i + ".wav");
			String[] cmdLine2 = {Shared.SOX_DIR + File.separator + "sox.exe", temporaryWavFile.getCanonicalPath(), wavFiles[i].getCanonicalPath(), "remix", "" + (i+1)};
			Process splitWav = Runtime.getRuntime().exec(cmdLine2);
			splitWav.waitFor();
			if (splitWav.exitValue() != 0) {
				throw new MusicFileException("Failed to split WAV files for " + this.getName());
			}
		}
		return wavFiles;
	}

}
