package us.lakora.brawl.rateconv.music;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.zip.GZIPInputStream;

import us.lakora.brawl.rateconv.MiniCopyFile;
import us.lakora.brawl.rateconv.ProgressBox;
import us.lakora.brawl.rateconv.Shared;

/*****************************************************************************
 *   Copyright 2012 Isaac Schemm
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   
 *   You may obtain a copy of the Apache License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 ****************************************************************************/

/**
 * This class is for VGM files, usually used for Sega music. Specs at: http://www.smspower.org/Music/VGMFileFormat
 */
public class VGM extends Music {

	public VGM(File input, int newRate) throws MusicFileException, IOException, InterruptedException {
		super(input);
		super.sampleRate = newRate;
		super.channels = 2;
		readHeader();
	}
	
	private void readHeader() throws IOException {
		// read four bytes
		int[] h = new int[4];
		FileInputStream temporary = new FileInputStream(input);
		for (int j=0; j<4; j++) {
			h[j] = temporary.read();
		}
		temporary.close();

		FileInputStream raw = new FileInputStream(input);
		InputStream is;
		if (h[0] == 0x1f && h[1] == 0x8b) {
			InputStream gzis = new GZIPInputStream(raw);
			is = gzis;
//			} else if (h[0] == 'V' && h[1] == 'g' && h[2] == 'm' && h[3] == ' ') {
		} else { // we can just assume that if it isn't GZIP, it's an uncompressed VGM file. If we didn't want to do that, we'd use the above check.
			is = raw;
		}
		
		// There aren't any checks for how many bytes are actually read, even though this value is returned by the read and skip methods. I find that this isn't necessary for a FileInputStream.
		is.skip(0x18);
		
		byte[] lengthB = new byte[4];
		is.read(lengthB, 0, 4);
		int length = 0;
		for (int j = 0; j < lengthB.length; j++)
		{
		   length += (lengthB[j] & 0xff) << (8 * j);
		}
		
		is.skip(0x4);

		byte[] loopSamplesB = new byte[4];
		is.read(loopSamplesB, 0, 4);
		int loopSamples = 0;
		for (int j = 0; j < loopSamplesB.length; j++)
		{
		   loopSamples += (loopSamplesB[j] & 0xff) << (8 * j);
		}

		int loopStart = -1;
		if (loopSamples != 0) {
			loopStart = length - loopSamples;
		}
		
		// for Brawl
		if (loopStart > -1) {
			int remainder = loopStart % 14336;
			if (remainder != 0) {
				loopStart += (14336 - remainder);
				length += (14336 - remainder);
			}
		}
		
		// In case the VGM was created with a sample rate besides 44100, we need to do the math here
		double ratio = (double)super.sampleRate / 44100;
		loopStart = (int)(loopStart * ratio);
		length = (int)(length * ratio);
		
		super.loopStart = loopStart;
		super.length = length;
	}
	
	public void createTemporaryWavFile(ProgressBox progressBox) throws IOException, InterruptedException, MusicFileException {
		progressBox.changeStatus("Dumping " + input.getName() + " to WAV...");
		
		// write INI file
		String ini = "[General]\n" +
				"SampleRate = " + super.sampleRate + "\n" +
				"LogSound = 1\n" +
				"MaxLoops = 0x02\n" +
				"MaxLoopsCMF = 0x01\n";
		FileWriter fw = new FileWriter(Shared.VGMPLAY_DIR + File.separator + "VGMPlay.ini");
		fw.write(ini);
		fw.close();
		
		File temporaryCopy = new File(Shared.TMP_DIR + File.separator + "vgmtmp.vgm");
		MiniCopyFile.copyFile(input, temporaryCopy);
		
		String[] cmdLine = {Shared.VGMPLAY_DIR + File.separator + "VGMPlay.exe", temporaryCopy.getCanonicalPath()};
		ProcessBuilder pb = new ProcessBuilder(cmdLine);
		Process dump = pb.start();
		final InputStream stdout = dump.getInputStream();

		final boolean debug = false;
		new Thread() {
			public void run() {
				try {
					InputStreamReader br = new InputStreamReader(stdout);
					int read = br.read();
					while (read > 0) {
						if (debug) System.err.print((char)read);
						read = stdout.read();
					}
				} catch (Exception e) {}
			}
		}.start();

		dump.waitFor();
		if (dump.exitValue() != 0) {
			throw new MusicFileException("VGMPlay.exe cannot read " + super.getCanonicalPath());
		}

		String path = temporaryCopy.getCanonicalPath();
		temporaryCopy.delete();
		String tempFileName = path.substring(0, path.length()-4) + ".wav";
		temporaryWavFile = new File(tempFileName);
		if (!temporaryWavFile.exists()) {
			throw new MusicFileException("The file " + tempFileName + " was put in the wrong folder.");
		}
		temporaryWavFile.deleteOnExit();
		
		progressBox.changeStatus("Converting " + input.getName() + "...");
	}

}
