package us.lakora.brawl.revb;

/**
 * RevB.java - Revolution B
 * Ported from C to Java by Isaac Schemm
 * Used under the Apache License 2.0 as part of Rate Converter by permission
 * from the original author (http://www.hcs64.com/vgm_ripping.html)
 */

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;

public class RevB {
	
	private static final byte[] RSTM_sig = {'R','S','T','M'};
	private static final byte[] head_name = {'H','E','A','D'};
	private static final byte[] adpc_name = {'A','D','P','C'};
	private static final byte[] data_name = {'D','A','T','A'};

	private static final int MODE_INVALID = 0;
	private static final int MODE_BUILD = 1;
	private static final int MODE_EXTRACT = 2;
	private static final int MODE_EXAMINE = 3;
	
	private static final int EXIT_FAILURE = 1;
	private static final int MAX_CHANNELS = 2;
	private static final String VERSION = "0.4j";
	
	private static String bin_name = null;
	
	public static class RevBException extends Exception {
		private static final long serialVersionUID = 1L;
		public RevBException(String message) {
			super(message);
		}
	}
	
	@SuppressWarnings("unused")
	private static void usage()
	{
	    System.out.printf(
	            "Revolution B for Java\n"+
	            "Version " + VERSION + "\n\n"+
	            "Build .brstm files from mono .dsp\n"+
	            "build usage:\n"+
	            "    %s --build dest.brstm source.dsp [sourceR.dsp%s] [options]\n"+
	            "build options:\n"+
	            "  --second-chunk-extra\n"+
	            "  --alternate-adpc-count\n",
	            bin_name, bin_name, (MAX_CHANNELS > 2 ? " ..." : ""), bin_name,
	            (MAX_CHANNELS > 2 ? " ..." : ""));

	    System.exit(1);
	}
	
	public static void main(String[] args) {
		final int argc = args.length+1;
		final String[] argv = new String[argc];
		argv[0] = "java -jar revb.jar";
		for (int i=1; i<argc; i++) argv[i] = args[i-1];
		
		boolean OPTION_SECOND_CHUNK_EXTRA = false;;
		boolean OPTION_ALTERNATE_ADPC_COUNT = false;

		String brstm_name = null;
		String[] dsp_names = new String[MAX_CHANNELS];
		int mode = MODE_INVALID;
		int dsp_count = 0;

		/* for usage() */
		bin_name = argv[0];

		/* clear dsp_names */
		for (int i = 0; i < MAX_CHANNELS; i++) dsp_names[i] = null;

		/* process arguments */
		for (int i = 1; i < argc; i++)
		{
			if ("--extract".equals(argv[i]))
			{
				if (mode != MODE_INVALID) usage();
				if (i >= argc-1) usage();

				mode = MODE_EXTRACT;
				brstm_name = argv[++i];
			}
			else if ("--build".equals(argv[i]))
			{
				if (mode != MODE_INVALID) usage();
				if (i >= argc-1) usage();

				mode = MODE_BUILD;
				brstm_name = argv[++i];
			}
			else if ("--examine".equals(argv[i]))
			{
				if (mode != MODE_INVALID) usage();
				if (i >= argc-1) usage();

				mode = MODE_EXAMINE;
				brstm_name = argv[++i];
			}
			else if ("--second-chunk-extra".equals(argv[i]))
			{
				if (OPTION_SECOND_CHUNK_EXTRA) usage();
				OPTION_SECOND_CHUNK_EXTRA = true;
			}
			else if ("--alternate-adpc-count".equals(argv[i]))
			{
				if (OPTION_ALTERNATE_ADPC_COUNT) usage();
				OPTION_ALTERNATE_ADPC_COUNT = true;
			}
			else
			{
				if (dsp_count == MAX_CHANNELS)
				{
					System.out.printf("Maximum of %d channels.\n",MAX_CHANNELS);
					System.exit(EXIT_FAILURE);
				}
				dsp_names[dsp_count++] = argv[i];
			}
		}

		/* some additional mode checks */
		if (mode != MODE_BUILD && (OPTION_ALTERNATE_ADPC_COUNT || OPTION_SECOND_CHUNK_EXTRA)) usage();
		if (mode == MODE_EXAMINE && dsp_count != 0) usage();
		if (mode != MODE_EXAMINE && dsp_count == 0) usage();

		try {
			switch (mode) {
			case MODE_BUILD:
				if (dsp_count > 2)
				{
					System.out.printf("Not comfortable with building > 2 channels\n");
					System.exit(EXIT_FAILURE);
				}
				build(brstm_name, dsp_names, dsp_count, OPTION_SECOND_CHUNK_EXTRA, OPTION_ALTERNATE_ADPC_COUNT);
				break;
			case MODE_EXAMINE:
			case MODE_EXTRACT:
				//	            extract(brstm_name, dsp_names, dsp_count);
				System.out.println("This mode is not implemented in the Java version.");
				break;
			case MODE_INVALID:
				usage();
				break;
			}
		} catch (FileNotFoundException e) {
			System.out.println("File not found: " + e.getMessage());
		} catch (Exception e) {
			e.printStackTrace();
			System.exit(EXIT_FAILURE);
		}

		System.exit(0);
	}
	
	public static void build(String brstm_name, String[] dsp_names,
			int dsp_count, boolean OPTION_SECOND_CHUNK_EXTRA, boolean OPTION_ALTERNATE_ADPC_COUNT) throws RevBException, IOException
	{
		File[] dsps = new File[dsp_count];
		for (int i=0; i<dsp_count; i++) {
			dsps[i] = new File(dsp_names[i]);
		}
		build(new File(brstm_name), dsps, dsp_count, OPTION_SECOND_CHUNK_EXTRA, OPTION_ALTERNATE_ADPC_COUNT);
	}

	public static void build(File brstm_name, File[] dsp_names,
			int dsp_count, boolean OPTION_SECOND_CHUNK_EXTRA, boolean OPTION_ALTERNATE_ADPC_COUNT) throws RevBException, IOException
	{
			for (int i=0; i<dsp_names.length; i++) {
				if (!dsp_names[i].exists()) throw new FileNotFoundException(dsp_names[i].getName());
			}
		
			RandomAccessFile outfile = null;
			RandomAccessFile[] infiles = new RandomAccessFile[MAX_CHANNELS];
	
//		    for (int i = 0; i < MAX_CHANNELS; i++)
//		    {
//		        infiles[i] = null;
//		    }
	
		    /* announce intentions */
		    System.out.printf("Building %s from:\n",brstm_name);
		    for (int i = 0; i < dsp_count; i++)
		    {
		        System.out.printf("  channel %d: %s\n", i, dsp_names[i]);
		    }
	
		    System.out.printf("\n");
	
		    /* open input files */
		    for (int i = 0; i < dsp_count; i++)
		    {
		        infiles[i] = new RandomAccessFile(dsp_names[i], "r");
		        check_errno(infiles[i] == null, "fopen of input file");
		    }
	
		    /* open output file */
		    outfile = new RandomAccessFile(brstm_name, "rw");
		    check_errno(outfile == null, "fopen of output file");
	
		    /* check that the DSPs agree */
		    byte[] dsp_header0 = new byte[0x60];
	
		    get_bytes_seek(0, infiles[0], dsp_header0, 0x1c);
		    for (int i = 1; i < dsp_count; i++)
		    {
		        byte[] dsp_header1 = new byte[0x1c];
		        get_bytes_seek(0, infiles[i], dsp_header1, 0x1c);
		        if (memcmp(dsp_header0, dsp_header1, 0x1c) != 0)
		        {
		            throw new RevBException("DSP headers do not agree");
		        }
		    }
	
		    /* read important elements of the header */
		    int nibble_count = read_32_be(dsp_header0, 4);
		    final int sample_rate = read_32_be(dsp_header0, 8);
		    final short loop_flag = read_16_be(dsp_header0, 0xc);
		    if (read_16_be(dsp_header0, 0xe) != 0)
		    {
		    	throw new RevBException("source file is not DSP ADPCM");
		    }
		    final int loop_nibble = read_32_be(dsp_header0, 0x10);
	
		    /* truncate to loop end (note that standard DSP loop end is the
		     * last nibble played, not one after) */
		    if ((loop_flag != 0) && read_32_be(dsp_header0, 0x14)+1 < nibble_count) {
		        nibble_count = read_32_be(dsp_header0, 0x14)+1;
		    }
		    final int sample_count = nibbles_to_samples(nibble_count);
	
		    /* now we can start building the file */
		    int current_offset = 0;
		    
		    /* begin RSTM header */
		    put_bytes_seek(current_offset, outfile, RSTM_sig, 4);
		    put_32_be(0xFEFF0100, outfile);
		    current_offset += 8;
		    /* reserve space for file size */
		    final int file_size_offset = current_offset;
		    current_offset += 4;
		    /* header size and ?? */
		    put_16_be_seek(0x40, current_offset, outfile);
		    put_16_be(0x2, outfile);
		    current_offset += 4;
		    /* reserve space for chunk info */
		    final int head_chunk_offset_offset = current_offset;
		    final int head_chunk_size_offset = current_offset + 4;
		    current_offset += 8;
		    final int adpc_chunk_offset_offset = current_offset;
		    final int adpc_chunk_size_offset = current_offset + 4;
		    current_offset += 8;
		    final int data_chunk_offset_offset = current_offset;
		    final int data_chunk_size_offset = current_offset + 4;
		    current_offset += 8;
	
		    /* pad */
		    current_offset = pad(current_offset, 0x20, outfile);
	
		    /* begin HEAD chunk header */
		    final int head_chunk_offset = current_offset;
		    put_32_be_seek(head_chunk_offset, head_chunk_offset_offset, outfile);
		    put_bytes_seek(current_offset, outfile, head_name, 4);
		    current_offset += 8;
	
		    /* HEAD body */
		    final int head_offset = current_offset;
		    final int block_size = 0x2000;
		    int last_block_size;
		    int last_block_used_bytes;
		    int block_count;
		    int head_data_offset_offset;
		    int samples_per_adpc_entry;
		    int bytes_per_adpc_entry;
		    {
		        /* chunk list */
		        put_32_be_seek(0x01000000, current_offset, outfile);
		        final int first_chunk_offset_offset = current_offset + 4;
		        current_offset += 8;
		        put_32_be_seek(0x01000000, current_offset, outfile);
		        final int second_chunk_offset_offset = current_offset + 4;
		        current_offset += 8;
		        put_32_be_seek(0x01000000, current_offset, outfile);
		        final int third_chunk_offset_offset = current_offset + 4;
		        current_offset += 8;
	
		        /* first chunk */
		        put_32_be_seek(current_offset-head_offset, first_chunk_offset_offset,
		                outfile);
		        put_byte_seek((byte)2, current_offset, outfile);  /* DSP ADPCM */
		        if (loop_flag == 0) {
		        	put_byte((byte)0,  outfile);
		        } else {
			        put_byte((byte)1,  outfile);
		        }
		        put_byte((byte)dsp_count, outfile);
		        put_byte((byte)0, outfile);   /* padding */
		        put_16_be(sample_rate, outfile);
		        put_16_be(0, outfile);  /* padding */
		        put_32_be(nibbles_to_samples(loop_nibble), outfile);    /* loop start */
		        put_32_be(sample_count, outfile);
		        current_offset += 0x10;
	
		        head_data_offset_offset = current_offset;
		        current_offset += 4;
	
		        /* DSP-centric */
		        final int samples_per_block = nibbles_to_samples(block_size*2);
		        if ( nibbles_to_samples(loop_nibble) % samples_per_block != 0 )
		        {
		            System.out.printf("Warning!\n"+
		                    "   Loop start sample %d is not on "+
		                    "a block boundary.\n"+
		                    "   The brstm may not loop properly "+
		                    "(blocks are %d samples).\n"+
		                    "   This can be solved by adding %d samples of\n"+
		                    "   silence to the beginning of the track.\n\n",
		                    nibbles_to_samples(loop_nibble), samples_per_block,
		                    samples_per_block -
		                    (nibbles_to_samples(loop_nibble) % samples_per_block)
		                    );
		        }
		        block_count = (sample_count + samples_per_block-1) / samples_per_block;
		        put_32_be_seek(block_count, current_offset, outfile);
		        put_32_be(block_size, outfile);
		        put_32_be(samples_per_block, outfile);
		        final int last_block_samples = sample_count % samples_per_block;
		        last_block_used_bytes = (samples_to_nibbles(last_block_samples)+1)/2;
		        last_block_size = (last_block_used_bytes + 0x1f) / 0x20 * 0x20;
		        put_32_be(last_block_used_bytes, outfile);
		        put_32_be(last_block_samples, outfile);
		        put_32_be(last_block_size, outfile);
	
		        if (OPTION_ALTERNATE_ADPC_COUNT)
		            samples_per_adpc_entry = 0x400;
		        else
		            samples_per_adpc_entry = samples_per_block;
		        put_32_be(samples_per_adpc_entry, outfile);
		        bytes_per_adpc_entry = 4;
		        put_32_be(bytes_per_adpc_entry, outfile);
		        current_offset += 0x20;
	
		        /* second chunk (don't know what this means) */
		        put_32_be_seek(current_offset-head_offset, second_chunk_offset_offset,
		                outfile);
		        if (OPTION_SECOND_CHUNK_EXTRA)
		        {
		            put_32_be_seek(0x01010000, current_offset, outfile);
		            put_32_be(0x01010000, outfile);
		            put_32_be(0x58, outfile);
		            put_32_be(0x7f400000, outfile);
		            put_32_be(0, outfile);
		            if (dsp_count == 1)
		                put_32_be(0x01000000, outfile);
		            else
		                put_32_be(0x02000100, outfile);
		            current_offset += 0x18;
		        }
		        else
		        {
		            put_32_be_seek(0x01000000, current_offset, outfile);
		            put_32_be(0x01000000, outfile);
		            put_32_be(0x58, outfile);
		            if (dsp_count == 1)
		                put_32_be(0x01000000, outfile);
		            else
		                put_32_be(0x02000100, outfile);
		            current_offset += 0x10;
		        }
	
		        /* third chunk */
		        put_32_be_seek(current_offset-head_offset, third_chunk_offset_offset,
		                outfile);
		        put_byte_seek((byte)dsp_count, current_offset, outfile);
		        byte[] padding = new byte[2];
		        put_bytes(outfile, padding, 2);   /* padding */
		        current_offset += 4;
		        int[] channel_info_offset_3x = new int[MAX_CHANNELS];
		        for (int i = 0; i < dsp_count; i++)
		        {
		            put_32_be_seek(0x01000000, current_offset, outfile);
		            current_offset+=4;
		            channel_info_offset_3x[i] = current_offset;
		            current_offset+=4;
		        }
		        int[] channel_info_offset_2x = new int[MAX_CHANNELS];
		        for (int i = 0; i < dsp_count; i++)
		        {
		            put_32_be_seek(current_offset-head_offset,
		                    channel_info_offset_3x[i], outfile);
		            put_32_be_seek(0x01000000, current_offset, outfile);
		            current_offset+=4;
		            channel_info_offset_2x[i] = current_offset;
		            current_offset+=4;
	
		            put_32_be_seek(current_offset-head_offset,
		                    channel_info_offset_2x[i], outfile);
		            byte[] dsp_header_channel_info = new byte[0x30];
		            get_bytes_seek(0x1c, infiles[i], dsp_header_channel_info, 0x30);
		            put_bytes(outfile, dsp_header_channel_info, 0x30);
		            current_offset += 0x30;
		        }
	
		        /* pad */
		        current_offset = pad(current_offset, 0x20, outfile);
	
		        /* done with HEAD chunk, store size */
		        put_32_be_seek(current_offset-head_chunk_offset,head_chunk_size_offset,
		                outfile);
		        put_32_be_seek(current_offset-head_chunk_offset,head_chunk_offset+4,
		                outfile);
		    } /* end write HEAD chunk */
	
		    /* begin ADPC chunk header */
		    final int adpc_chunk_offset = current_offset;
		    put_32_be_seek(adpc_chunk_offset, adpc_chunk_offset_offset, outfile);
		    put_bytes_seek(current_offset, outfile, adpc_name, 4);
		    current_offset += 8;
	
		    /* ADPC body */
		    {
		        int adpc_blocks;
		        if (OPTION_ALTERNATE_ADPC_COUNT)
		        {
		            adpc_blocks = (block_size * (block_count-1) + last_block_size) /
		                0x400 + 1;
		        }
		        else
		        {
		            adpc_blocks = sample_count / samples_per_adpc_entry + 1;
		        	}
	
		        /* TODO actually fill in ADPC values */
		        current_offset +=
		            adpc_blocks * bytes_per_adpc_entry * dsp_count;
	
		        /* pad */
		        current_offset = pad(current_offset, 0x20, outfile);
	
		        /* done with ADPC chunk, store size */
		        put_32_be_seek(current_offset-adpc_chunk_offset,adpc_chunk_size_offset,
		                outfile);
		        put_32_be_seek(current_offset-adpc_chunk_offset,adpc_chunk_offset+4,
		                outfile);
		    } /* end write ADPC chunk */
	
		    /* begin DATA chunk header */
		    final int data_chunk_offset = current_offset;
		    put_32_be_seek(data_chunk_offset, data_chunk_offset_offset, outfile);
		    put_bytes_seek(current_offset, outfile, data_name, 4);
		    current_offset += 8;
	
		    /* DATA body */
		    {
		        int infile_offset = 0x60;
		        outfile.seek(current_offset);
	
		        put_32_be(0x18, outfile);
		        current_offset += 0x18;
		        put_32_be_seek(current_offset, head_data_offset_offset, outfile);
		        outfile.seek(current_offset);
	
		        for (int i = 0; i < block_count; i++)
		        {
		            if (i == block_count-1 && last_block_size != 0) break;
	
		            for (int channel = 0; channel < dsp_count; channel++)
		            {
		                dump(infiles[channel], outfile, infile_offset, block_size);
		                current_offset += block_size;
		            }
	
		            infile_offset += block_size;
		        }
		        if (last_block_size != 0)
		        {
		            for (int channel = 0; channel < dsp_count; channel++)
		            {
		                outfile.seek(current_offset);
		                dump(infiles[channel], outfile, infile_offset,
		                        last_block_used_bytes);
	
		                /* pad */
		                for (int i = last_block_used_bytes; i < last_block_size;
		                        i++)
		                {
		                    put_byte((byte)0, outfile);
		                }
		                current_offset += last_block_size;
		            }
		        }
	
		        /* pad */
		        current_offset = pad(current_offset, 0x20, outfile);
	
		        /* done with DATA chunk, store size */
		        put_32_be_seek(current_offset-data_chunk_offset,data_chunk_size_offset,
		                outfile);
		        put_32_be_seek(current_offset-data_chunk_offset,data_chunk_offset+4,
		                outfile);
		    } /* end write DATA chunk */
	
		    /* done with file */
		    put_32_be_seek(current_offset, file_size_offset, outfile);
	
		    /* close files */
		    for (int i = 0; i < dsp_count; i++)
		    {
		        infiles[i].close();
		        infiles[i] = null;
		    }
		    outfile.close();
		    outfile = null;
	
		    System.out.printf("Done!\n");
	}

	private static int memcmp(byte[] b1, byte[] b2, int length) {
		int i = 0;
		while (i < length) {
			int difference = b1[i] - b2[i];
			if (difference != 0) {
				return difference;
			}
			i++;
		}
		return 0;
	}
	
	private static int unsignedByteToInt(byte b) {
		int r = b;
    	if (r < 0) {
    		r += 256;
    	}
    	return r;
	}
	
	private static void check_errno(boolean condition, String message) throws RevBException {
		if (condition) throw new RevBException(message);
	}
	private static void check_errno(boolean condition, Object o, String message) throws RevBException {
		check_errno(condition, o + " " + message);
	}
	
	private static void get_bytes(RandomAccessFile f, byte[] buf, int byte_count) throws IOException, RevBException
	{
		int bytes_read = 0;
		bytes_read = f.read(buf, 0, byte_count);
		check_errno(bytes_read != byte_count, f, "fread");
	}

	private static void get_bytes_seek(long offset, RandomAccessFile infile, byte[] buf, int byte_count) throws IOException, RevBException
	{
		infile.seek(offset);
	    get_bytes(infile, buf, byte_count);
	}
	
	private static void put_bytes(RandomAccessFile infile, byte[] buf, int byte_count) throws IOException
	{
		infile.write(buf, 0, byte_count);
	}

	private static void put_bytes_seek(long offset, RandomAccessFile infile, byte[] buf, int byte_count) throws IOException
	{
	    infile.seek(offset);
	    put_bytes(infile, buf, byte_count);
	}
	
	private static void put_32_be(int value, RandomAccessFile infile) throws IOException
	{
	    byte[] buf = new byte[4];
	    write_32_be(value, buf);
	    infile.write(buf);
	}
	private static void put_32_be_seek(int value, long offset, RandomAccessFile infile) throws IOException
	{
		infile.seek(offset);
	    put_32_be(value, infile);
	}

	private static void put_16_be(int value, RandomAccessFile infile) throws IOException
	{
	    byte[] buf = new byte[2];
	    write_16_be((short)value, buf);
	    infile.write(buf);
	}
	private static void put_16_be_seek(int value, long offset, RandomAccessFile infile) throws IOException
	{
		infile.seek(offset);
	    put_16_be(value, infile);
	}
	
	private static void put_byte(byte value, RandomAccessFile infile) throws IOException
	{
	    byte[] buf = {value};
	    infile.write(buf);
	}
	private static void put_byte_seek(byte value, long offset, RandomAccessFile infile) throws IOException
	{
	    infile.seek(offset);
	    put_byte(value, infile);
	}
	
	private static void dump(RandomAccessFile infile, RandomAccessFile outfile, int offset, int size) throws IOException {
		infile.seek(offset);
		byte[] data = new byte[size];
		infile.read(data);
		outfile.write(data);
	}
	
//	private static int read_32_le(byte[] bytes, int start) {
//		int result = 0;
//		for (int i=3; i>=0; i--) result = (result << 8) | unsignedByteToInt(bytes[i+start]);
//		return result;
//	}
//	private static short read_16_le(byte[] bytes, int start)
//	{
//	    int result = 0;
//	    for (int i=1; i>=0; i--) result = (result << 8) | unsignedByteToInt(bytes[i+start]);
//	    return (short)result;
//	}
//	private static long read_64_be(byte[] bytes, int start)
//	{
//	    long result = 0;
//	    for (int i=0; i<8; i++) result = (result << 8) | unsignedByteToInt(bytes[i+start]);
//	    return result;
//	}
	private static int read_32_be(byte[] bytes, int start)
	{
	    int result = 0;
	    for (int i=0; i<4; i++) result = (result << 8) | unsignedByteToInt(bytes[i+start]);
	    return result;
	}
	private static short read_16_be(byte[] bytes, int start)
	{
	    int result = 0;
	    for (int i=0; i<2; i++) result = (result << 8) | unsignedByteToInt(bytes[i+start]);
	    return (short)result;
	}

	private static byte[] write_32_be(int value, byte[] bytes)
	{
	    for (int i=3; i>=0; i--, value >>= 8) bytes[i] = (byte) (value & 0xff);
	    return bytes;
	}

	private static byte[] write_16_be(short value, byte[] bytes)
	{
	    for (int i=1; i>=0; i--, value >>= 8) bytes[i] = (byte) (value & 0xff);
	    return bytes;
	}
	
	private static int samples_to_nibbles(int samples)
	{
	    int nibbles = samples / 14 * 16;
	    if (samples % 14 != 0) nibbles += 2 + samples % 14;
	    return nibbles;
	}

	private static int nibbles_to_samples(int nibbles)
	{
	    int whole_frames = nibbles / 16;
	    if (nibbles % 16 != 0)
	    {
	        return whole_frames * 14 + nibbles % 16 - 2;
	    }
	    return whole_frames * 14;
	}
	
	private static int pad(int current_offset, int pad_amount, RandomAccessFile outfile) throws IOException
	{
	    int new_offset = (current_offset + pad_amount-1) / pad_amount * pad_amount;

	    for (; current_offset < new_offset; current_offset++)
	    {
	        put_byte_seek((byte)0, current_offset, outfile);
	    }

	    return new_offset;
	}

}
