package us.lakora.dialoglist;

/*****************************************************************************
 *   Copyright 2011 Isaac Schemm
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   
 *   You may obtain a copy of the Apache License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 ****************************************************************************/

import java.util.ArrayList;

import javax.swing.JButton;
import javax.swing.JOptionPane;

/**
 * A class to keep track of a list of dialogs (useful for large help sections) with Back, Next and Close buttons.
 * Last updated: January 8, 2011
 * @author Isaac Schemm
 */
public class DialogList {

	private ArrayList<Object> messageList;
	private ArrayList<Integer> messageTypes;
	private String title;
	private boolean closeButton;
	
	private String backName;
	private String closeName;
	private String nextName;
	
	/**
	 * Constructs a DialogList with no messages.<br/>
	 * Running {@link DialogList#showDialog showDialog} before adding at least one message to display will result in an {@link IndexOutOfBoundsException}.
	 */
	public DialogList() {
		messageList = new ArrayList<Object>();
		messageTypes = new ArrayList<Integer>();
		title = "";
		closeButton = true;
		backName = "Back";
		closeName = "Close";
		nextName = "Next";
	}
	
	/**
	 * Sets whether the message dialogs will show a Close button between Back and Next when presented to the user.
	 * @param b determines whether a Close button will appear. The default is <code>true</code>.
	 */
	public void setCloseButtonEnabled(boolean b) {
		closeButton = b;
	}
	
	/**
	 * Adds a message dialog to the list, of the type {@link JOptionPane.PLAIN_MESSAGE}.
	 * @param o The object to display as the message. Usually a {@link String}, but it can also be a {@link java.awt.Component Component}.
	 * @return <code>true</code> (as specified by {@link java.util.Collection#add Collection.add(E)}, from {@link ArrayList})
	 */
	public boolean add(Object o) {
		messageTypes.add(JOptionPane.PLAIN_MESSAGE);
		return messageList.add(o);
	}
	
	/**
	 * Adds a message dialog to the list, of the given type.
	 * @param o The object to display as the message. Usually a {@link String}, but it can also be a {@link java.awt.Component Component}.
	 * @param messageType A type for {@link JOptionPane}. Available types are {@link JOptionPane#ERROR_MESSAGE},
	 * {@link JOptionPane#INFORMATION_MESSAGE}, {@link JOptionPane#WARNING_MESSAGE},
	 * {@link JOptionPane#QUESTION_MESSAGE}, and {@link JOptionPane#PLAIN_MESSAGE}.
	 * @return <code>true</code> (as specified by {@link java.util.Collection#add Collection.add(E)}, from {@link ArrayList})
	 */
	public boolean add(Object o, int messageType) {
		messageTypes.add(messageType);
		return messageList.add(o);
	}
	
	/**
	 * Removes the message dialog at the specified position in this list. Shifts any subsequent elements to the left (subtracts one from their indices).
	 * @param index the index of the element to be removed
	 * @return the element that was removed from the list
	 */
	public Object remove(int index) {
		messageTypes.remove(index);
		return messageList.remove(index);
	}
	
	/**
	 * Removes the first occurrence of the specified message dialog from this list, if it is present, using {@link ArrayList#remove(Object)}.
	 * <br/>The corresponding message type will also be removed from the list.
	 * @param o element to search for
	 * @return the result of {@link ArrayList#remove(Object)} on the list containing the messages
	 */
	public boolean remove(Object o) {
		messageTypes.remove(messageList.indexOf(o));
		return messageList.remove(o);
	}
	
	/**
	 * Returns the message at the specified position in this list. This might be useful if the message is a {@link java.awt.Component Component} with configurable properties.
	 * @param index index of the element to return
	 * @return the element at the specified position in this list
	 */
	public Object get(int index) {
		return messageList.get(index);
	}
	
	/**
	 * Sets the title that each dialog will display to the specified string.
	 * @param s the string to use in the title bar. The default is an empty string.
	 */
	public void setTitle(String s) {
		title = s;
	}
	
	/**
	 * Sets the text on the "Back" button.
	 * @param s the string to use. The default is "Back".
	 */
	public void setBackButtonTitle(String s) {
		backName = s;
	}
	
	/**
	 * Sets the text on the "Close" button.
	 * @param s the string to use. The default is "Close".
	 */
	public void setCloseButtonTitle(String s) {
		closeName = s;
	}
	
	/**
	 * Sets the text on the "Next" button.
	 * @param s the string to use. The default is "Next".
	 */
	public void setNextButtonTitle(String s) {
		nextName = s;
	}
	
	/**
	 * @return a string representing each message in the list (obtained via each object's <code>toString</code> method), separated by two newlines.
	 * <br/>If there are no messages in the list, an empty string is returned.
	 */
	public String toString() {
		String s;
		if (messageList.size() < 1) {
			s = "";
		} else {
			s = messageList.get(0).toString();
		}
		for (int i=1; i<messageList.size(); i++) {
			s = s + "\n\n" + messageList.get(i).toString();
		}
		return s;
	}
	
	public boolean equals(Object o) {
		DialogList dl = (DialogList)o;
		return (messageList.equals(dl.messageList));
	}
	
	public int size() {
		return messageList.size();
	}
	
	/**
	 * Shows the message dialog at the specified position in the list.
	 * @param index the index of the message dialog to start at
	 * @return the index of the last message dialog shown when the user clicked Close or closed the window
	 * @throws IndexOutOfBoundsException if the index is below 0 or is not less than the number of elements in the list
	 */
	public int showDialog(int index) throws IndexOutOfBoundsException {
		if ((index < 0) || (index >= messageList.size())) {
			throw new IndexOutOfBoundsException();
		} else {
			Object[] opts;
			int lastOpt;
			if (closeButton) {
				opts = new Object[3];
				opts[1] = closeName;
				lastOpt = 2;
			} else {
				opts = new Object[2];
				lastOpt = 1;
			}
			if (index == 0) {
				JButton b = new JButton();
				b.setText(backName);
				b.setEnabled(false);
				opts[0] = b;
			} else {
				opts[0] = backName;
			}
			if (index == messageList.size()-1) {
				JButton b = new JButton();
				b.setText(nextName);
				b.setEnabled(false);
				opts[lastOpt] = b;
			} else {
				opts[lastOpt] = nextName;
			}
			int ret = JOptionPane.showOptionDialog(null, messageList.get(index), title,
					JOptionPane.OK_CANCEL_OPTION, messageTypes.get(index),
					null, opts, null);
			int finalRet;
			if (ret == 0) {
				finalRet = showDialog(index-1);
			} else if (ret == lastOpt) {
				finalRet = showDialog(index+1);
			} else {
				finalRet = index;
			}
			return finalRet;
		}
	}
}
